
!Physical variables
MODULE FIELD
IMPLICIT NONE
SAVE

INTEGER,PARAMETER::NINP=21,ANIMATION=22, RUNDATA=23, INTERP=24
INTEGER::IBAR,JBAR,IM1,JM1,IMAX,JMAX,I,J,NFRAME,NTHFRAME,&
UCOUNT,UCOUNT_T,ENTRIESA,FLAG,PREFNODE,ENTRIESA_T
DOUBLE PRECISION::DELT,T,TMAX,DELX,DELY,RELAX,XMAX,YMAX,RDX,RDY,&
TH,TC,TM,ADVX,ADVY,DIFFX,DIFFY,PR,RA,STE,VF,&
A,INITA,MC,STC,FB,TH_RE,TM_RE,TC_RE,THERM_EXP,T_MID,&
CON_A,CON_B,CON_C,CON_D,CON_E,H,G,TAU,FSLIM,RHO_GAL1,RHO_GAL2,&
CAP_GAL1,CAP_GAL2,KL_GAL1,KL_GAL2,CPU_START,CPU_STOP,DELTN
DOUBLE PRECISION,ALLOCATABLE,DIMENSION(:,:)::U,V,P,X,Y,TEMP,TEMP_P,FS,&
FS_P,FS_N,PSI,XI,XI_N,KON
INTEGER, ALLOCATABLE, DIMENSION (:,:) :: NN,NNT,FNP,GEO
DOUBLE PRECISION, ALLOCATABLE, DIMENSION (:) :: B,DELP,AVAL,PX,PY,TB,TAVAL,TEMP_VEC
INTEGER, ALLOCATABLE, DIMENSION (:) :: AI,AJ,TI,TJ

END MODULE FIELD

!Pressure solver variables
MODULE SOLVEFIELD
IMPLICIT NONE
SAVE

INTEGER :: ITER_P,ITER_T,ITER_TT,ITMAX_TT
DOUBLE PRECISION :: RES_P,RES_T,RES_TT,EPSI_TT

END MODULE SOLVEFIELD




PROGRAM CAVSOL2D
!-------------------------------------------------------------
!     SOLVES THE NAVIER-STOKES EQUATIONS ON A 2-DIMENSIONAL
!     RECTANGULAR GRID WITH A SOLIDIFYING FLUID
!     Version 21.april 2002 Reidar Kristoffersen
!     Version 01.november 2017 Olav Rømcke
!------------------------------------------------------------
	



USE FIELD
USE SOLVEFIELD

IMPLICIT NONE


CALL INPUT                ! Read input
CALL DEF_GEO		  ! Defining geometry
CALL PREPARE_TECPLOT_FILE ! Prepare files for writing data
CALL INITVPT  		  ! Define initial Thermal and momentum fields.
CALL BCVEL    		  ! Impose boundary conditions by calling BCVEL.
CALL BCTEMP   		  ! Impose boundary conditions by calling BCTEMP.
CALL NODE_NUMBERS_TEMP	  ! define node numbers for temperature field


WRITE(*,*)'PR =',PR,' RA = ',RA,' STE = ',STE,' TAU = ',TAU,' TH = ',TH_RE,&
' TM = ',TM_RE,' TC = ',TC_RE,' KL = ',KL_GAL1,' KS = ',KL_GAL2,' CL = ',&
CAP_GAL1,' CS = ',CAP_GAL2,' RL = ',RHO_GAL1,' RS = ',RHO_GAL2,' H = ',H





!     ==== START OF GLOBAL LOOP ====

NTHFRAME = CEILING(DELTN/DELT)		! Capture physics every nth timestep
NFRAME = 0

CALL CPU_TIME(CPU_START)

  
DO WHILE(.TRUE.)
  IF(MOD(NFRAME,NTHFRAME).EQ.0) THEN
    CALL WRITE_TO_TECPLOT
  ENDIF
  NFRAME = NFRAME + 1
  CALL WRITE_TO_TECPLOT2
  T=T+DELT
  CALL DEF_NODE 
  CALL NODE_NUMBERS
  CALL PITER
  CALL TITER
  CALL CONSERVATION
  CALL PHYS_VAL
 
  WRITE(*,*)
  WRITE(*,*)
  WRITE(*,'(A,G12.5)') ' Time = ',T
  WRITE(*,*)'-------------------------------------------------------------------------------------'
  WRITE(*,'(3(A,I12))')  '  TT-iter = ',ITER_TT,' |  T-iter = ',ITER_T,' |  P-iter = ',ITER_P
  WRITE(*,'(3(A,G12.5))')'  TT-res  = ',RES_TT, ' |  T-res  = ',RES_T, ' |  P-res  = ',RES_P
  WRITE(*,'(2(A,G12.5))')'  FS-glob = ',VF,     ' |  MC-glob = ',MC 
  WRITE(*,*)'-------------------------------------------------------------------------------------'

  IF(T.GE.TMAX)EXIT
END DO


!     ==== END OF GLOBAL LOOP ====

CALL CPU_TIME(CPU_STOP)


CALL WRITE_TO_TECPLOT
CALL WRITE_TO_TECPLOT2






! end routine
WRITE(*,*)'Simulation finished.'
WRITE(*,*)'PR =',PR,' RA = ',RA,' STE = ',STE,' TAU = ',TAU
WRITE(*,*)'Duration: ', CPU_STOP-CPU_START, ' sec'

DEALLOCATE (U,V,P,TEMP,TEMP_P,X,Y,FS,FS_P,PSI,NN,PX,PY,FNP,&
NNT,TAVAL,TI,TJ,TB,TEMP_VEC,KON)

STOP
END PROGRAM

!------------------------------------------------------------

SUBROUTINE INPUT
!     purpose:
!     To read input data and compute parameters from input file 
!     The routine also checks several stability criteria.
USE FIELD
USE SOLVEFIELD
IMPLICIT NONE
INTEGER ::IFAIL
DOUBLE PRECISION ::DUMMY,TEST,TESTT,TESTTT,TEST1,TEST2,TESTTTT

OPEN (UNIT=NINP,FILE='SIMINPUT2D.INP',FORM='FORMATTED')
READ(NINP,1000)
READ(NINP,1020)TMAX,DELT,DELTN,DELX,DELY,XMAX,YMAX,RA,PR,STE,TH_RE,&
TM_RE,TC_RE,KL_GAL1,KL_GAL2,CAP_GAL1,CAP_GAL2,RHO_GAL1,RHO_GAL2,H


IBAR=NINT(DELX)      !define grid
JBAR=NINT(DELY)
DELX=XMAX/DELX
DELY=YMAX/DELY
IM1=IBAR+1
JM1=JBAR+1
IMAX=IBAR+2
JMAX=JBAR+2
INITA=XMAX*YMAX
A=INITA


!     ===== Check stability criteria =====

!  CODE HERE

!   Initiating parameters and fields

RDX  = 1.0/DELX
RDY  = 1.0/DELY
!set time to zero initially
T=0.0

! Time constant	  [s]
TAU = (H**2*RHO_GAL1*CAP_GAL1/KL_GAL1)

!set mass conservation to zero initially
MC=0

! solid liquid interphase limit
FSLIM=0.5

! Temperature scale
TH=(TH_RE-TC_RE)/(TH_RE-TC_RE)
TC=(TC_RE-TC_RE)/(TH_RE-TC_RE)
TM=(TM_RE-TC_RE)/(TH_RE-TC_RE)

! entries TA (temperature matrix)
	    ! in domain		  ! walls	      ! corners
ENTRIESA_T = 5*(IBAR-2)*(JBAR-2) + 4*2*(IBAR+JBAR-4) + 3*4

ALLOCATE (U(IBAR+2,JBAR+2),V(IBAR+2,JBAR+2),&
P(IBAR+2,JBAR+2),TEMP(IBAR+2,JBAR+2),TEMP_P(IBAR+2,JBAR+2),&
KON(IBAR+2,JBAR+2),X(IBAR+2,JBAR+2),Y(IBAR+2,JBAR+2),FS(IBAR+2,JBAR+2),&
FS_P(IBAR+2,JBAR+2),FS_N(IBAR+2,JBAR+2),PSI(IBAR+2,JBAR+2),XI(IBAR+2,JBAR+2),&
XI_N(IBAR+2,JBAR+2),NN(IBAR+2,JBAR+2),GEO(IBAR+2,JBAR+2),PX(JBAR),PY(JBAR),&
FNP(IBAR+2,JBAR+2),NNT(IBAR+2,JBAR+2),TAVAL(ENTRIESA_T),TI(ENTRIESA_T),&
TJ(ENTRIESA_T),TB(IBAR*JBAR),TEMP_VEC(IBAR*JBAR))

1000 FORMAT(//////)
1020 FORMAT(10(43X,F20.0/))
RETURN
END 

!------------------------------------------------------------

SUBROUTINE PREPARE_TECPLOT_FILE
USE FIELD
IMPLICIT NONE


DO I = 2,IM1
  DO J = 2,JM1
    X(I,J) = (I-1.5)*DELX
    Y(I,J) = (J-1.5)*DELY
  ENDDO
ENDDO

OPEN (UNIT=ANIMATION,FILE='TECPLOT.DAT',FORM='FORMATTED')  ! Output file
WRITE(ANIMATION,*)'TITLE     = "Animate"'
WRITE(ANIMATION,*)'VARIABLES = "X","Y","U","V","P","PSI","T","FS","K","NN","NNT","FP",GEO'
WRITE(ANIMATION,*)''

OPEN (UNIT=RUNDATA,FILE='TECPLOT2.DAT',FORM='FORMATTED')  ! Output file
WRITE(RUNDATA,*)'TITLE     = "Mass conservation"'
WRITE(RUNDATA,*)'VARIABLES = "T", "MC", "VF", "RES_P", "ITER_P", "RES_T", "ITER_T", "RES_TT", "ITER_TT"'
WRITE(RUNDATA,*)''

OPEN (UNIT=INTERP,FILE='TECPLOT3.DAT',FORM='FORMATTED')  ! Output file
WRITE(INTERP,*)'TITLE     = "Inter Phase"'
WRITE(INTERP,*)'VARIABLES = "X", "Y"'
WRITE(INTERP,*)''


RETURN
END 



!------------------------------------------------------------
SUBROUTINE DEF_GEO
USE FIELD
IMPLICIT NONE

DOUBLE PRECISION :: VMA, VMI, VMAMA,VMIMI, HMA, HMI,HMAMA,HMIMI
GEO(:,:)=0

! Vertical borderlines
VMAMA=5/6.0
VMIMI=1/6.0
VMA=4/6.0
VMI=2/6.0

! Horisontal borderlines
HMAMA=5/6.0
HMIMI=1/6.0
HMA=4/6.0
HMI=2/6.0

DO I = 1,IMAX
  DO J = 1,JMAX

     !Indent
     IF ((J.GT.CEILING(JMAX*HMI)).AND.(J.LT.CEILING(JMAX*HMA))) THEN
	! Left	
	IF (I.LT.CEILING(IMAX*VMI)) THEN
	   GEO(I,J)=1
	! Right
	ELSEIF (I.GT.CEILING(IMAX*VMA)) THEN
	   GEO(I,J)=1
	ENDIF

     ENDIF

	! Top/Bottom mould
     IF ((J.GT.CEILING(JMAX*HMAMA)).OR.(J.LT.CEILING(JMAX*HMIMI))) THEN
	   GEO(I,J)=1
     ENDIF

	! Right/left mould
     IF ((I.GT.CEILING(IMAX*VMAMA)).OR.(I.LT.CEILING(IMAX*VMIMI))) THEN
	   GEO(I,J)=1
     ENDIF

  ENDDO
ENDDO



RETURN
END 

!------------------------------------------------------------

SUBROUTINE WRITE_TO_TECPLOT
!     purpose:
!     To create a dat file for use in tecplot
!     Will be used to create an animation
USE FIELD
IMPLICIT NONE

!Calculate "streamfunction"
PSI(:,:)=0.0

DO I=2,IM1
  PSI(I,1)=PSI(I-1,1) - V(I,1)*DELX
ENDDO

DO I=1,IM1
  DO J=2,JM1
     PSI(I,J)=PSI(I,J-1) + U(I,J)*DELY
  ENDDO
ENDDO



! Calculate solid-liquid Phase front x-y coordinates
PX(:)=0.0
PY(:)=0.0

DO J=2,JM1
  PY(J-1)=Y(2,J)
  DO I=2,IM1
     IF ((FS(I,J).GT.(FSLIM)).AND.(FS(I+1,J).LE.(FSLIM))) THEN
	PX(J-1)=X(I,J) + DELX*(FSLIM-FS(I,J))/(FS(I+1,J)-FS(I,J))
     ENDIF
     IF ((FS(I,J).LE.(FSLIM)).AND.(FS(I+1,J).GT.(FSLIM))) THEN
	PX(J-1)=X(I,J) - DELX*(FS(I+1,J)-FSLIM)/(FS(I+1,J)-FS(I,J))
     ENDIF
     
  ENDDO
ENDDO


! Write to ANIMATION file
WRITE(ANIMATION,*)'ZONE ','T="time = ',T,'" I=',(IM1-1),', J=',(JM1-1),', DATAPACKING=POINT'
DO J=2,JM1
  DO I=2,IM1
    WRITE(ANIMATION,*)X(I,J),' ',Y(I,J),' ',((U(I-1,J)+U(I,J))/2),' ',((V(I,J-1)+V(I,J))/2),' ',P(I,J),&
' ',PSI(I,J),' ',TEMP(I,J),' ',FS(I,J),' ',KON(I,J),' ',NN(I,J),' ',NNT(I,J),' ',FNP(I,J),' ',GEO(I,J)
  ENDDO
ENDDO
WRITE(ANIMATION,*)


! Write to INTERPHASE file
WRITE(INTERP,*)'ZONE ','T="time = ',T,'" J=',(JM1-1),', DATAPACKING=POINT'
DO J=1,JM1-1
    WRITE(INTERP,*)PX(J),' ',PY(J)
ENDDO
WRITE(INTERP,*)


RETURN
END 

!------------------------------------------------------------

SUBROUTINE WRITE_TO_TECPLOT2
!     purpose:
!     To create a dat file for use in tecplot
!     Will be used to create an animation
USE FIELD
USE SOLVEFIELD
IMPLICIT NONE


VF=0.0

DO I=2,IM1
  DO J=2,JM1
    VF = VF + FS(I,J)*DELX*DELY/(XMAX*YMAX)
  ENDDO
ENDDO


! Write to run data file
WRITE(RUNDATA,*)T,' ',MC,' ',VF,' ',RES_P,' ',ITER_P,' ',RES_T,' ',ITER_T,' ',RES_TT,' ',ITER_TT



RETURN
END 

!------------------------------------------------------------

SUBROUTINE INITVPT
!     Purpose:
!     To initiate the Thermal and Momentum fields 
USE FIELD
IMPLICIT NONE

DO J=1,JMAX
  DO I=1,IMAX
    U(I,J)=0.0
    V(I,J)=0.0
    P(I,J)=0.0
    IF (GEO(I,J).EQ.1) THEN
    	TEMP(I,J)=TC
    ELSE
	TEMP(I,J)=TH
    ENDIF

    FS(I,J)=0.0
    KON(I,J)=1.0 + (KL_GAL2-KL_GAL1)/KL_GAL1*GEO(I,J)
  ENDDO
ENDDO 
RETURN
END


!------------------------------------------------------------

SUBROUTINE PHYS_VAL
!     Purpose:
!     To initiate the Thermal and Momentum fields 
USE FIELD
IMPLICIT NONE

DO J=1,JMAX
  DO I=1,IMAX
    KON(I,J)=1.0 + (KL_GAL2-KL_GAL1)/KL_GAL1*GEO(I,J)
  ENDDO
ENDDO 
RETURN
END

!------------------------------------------------------------

SUBROUTINE NODE_NUMBERS_TEMP
!     purpose:
!     To assign node numbers to thermal fields
!     UCOUNT_T = number of unknowns
!     NNT(i,j) = node number associated with node at position (i,j)
USE FIELD
IMPLICIT NONE

NNT(:,:) = 0
UCOUNT_T = 0		

DO I=2,IM1
  DO J=2,JM1
    UCOUNT_T = UCOUNT_T + 1
    NNT(I,J) = UCOUNT_T
  ENDDO
ENDDO


RETURN
END

!------------------------------------------------------------

SUBROUTINE NODE_NUMBERS
!     purpose:
!     To assign node numbers to momentum fields
!     UCOUNT = number of unknowns, 
!     NN(i,j) = node number associated with node at position (i,j)
USE FIELD
IMPLICIT NONE

NN(:,:) = 0
UCOUNT = 0		

DO I=2,IM1
  DO J=2,JM1
    IF (FNP(I,J).EQ.1) THEN 
      UCOUNT = UCOUNT + 1
      NN(I,J) = UCOUNT
    ENDIF
  ENDDO
ENDDO



RETURN
END

!------------------------------------------------------------

SUBROUTINE DEF_NODE
!     purpose:
!     To assign node type, Solid node or fluid node
USE FIELD
IMPLICIT NONE

FNP(:,:)=0


!Fluid nodes
DO I=2,IM1
  DO J=2,JM1
    !Pressure cell
    IF ((FS(I,J).LT.FSLIM).AND.(GEO(I,J).EQ.0)) THEN
       FNP(I,J)=1
    ENDIF
  ENDDO
ENDDO



RETURN
END

!------------------------------------------------------------

SUBROUTINE CONSERVATION
!     purpose:
!     To find the area
USE FIELD
IMPLICIT NONE


DO I=2,IM1
  DO J=2,JM1
    MC = MC + DELT*((U(I,J)-U(I-1,J))*DELY+(V(I,J)-V(I,J-1))*DELX)/(XMAX*YMAX)
  ENDDO
ENDDO



RETURN
END

!------------------------------------------------------------

SUBROUTINE BCVEL
!     purpose:
!     To give boundary conditions to the velocities around the domain
USE FIELD
IMPLICIT NONE
DO I=1,IM1
  DO J=1,JM1
    IF((FNP(I,J).EQ.0).OR.(FNP(I+1,J).EQ.0)) THEN
      U(I,J) = 0.0
    ENDIF
    IF((FNP(I,J).EQ.0).OR.(FNP(I,J+1).EQ.0)) THEN
      V(I,J) = 0.0
    ENDIF
  ENDDO
ENDDO
RETURN
END

!------------------------------------------------------------

SUBROUTINE BCP
!     purpose:
!     To give boundary conditions to the Pressure around the domain
USE FIELD
IMPLICIT NONE


DO J=1,JMAX               !Impose boundary condition in y-direction
  !Left
  P(1,J)=P(2,J)
  !Right
  P(IM1,J) = P(IMAX,J)
ENDDO 

DO I=1,IMAX               !Impose boudary condition in x-direction
  !Bottom
  P(I,1)=P(I,2)
  !Top
  P(I,JMAX)=P(I,JM1)
ENDDO   


RETURN
END

!------------------------------------------------------------

SUBROUTINE BCTEMP
!     purpose:
!     To give boundary conditions to the temp and fs
USE FIELD
IMPLICIT NONE

DO J=1,JMAX               !Impose boudary condition in y-direction
  TEMP(1,J) = 2*TC-TEMP(2,J);
  TEMP(IMAX,J) = 2*TC-TEMP(IMAX-1,J);
  FS(1,J)=FS(2,J);
  FS(IMAX,J)=FS(IMAX-1,J);
ENDDO

DO I=1,IMAX                !Impose boudary condition in x-direction
  TEMP(I,1) = 2*TC - TEMP(I,2);
  TEMP(I,JMAX) = 2*TC - TEMP(I,JMAX-1);
  FS(I,1)=FS(I,2);
  FS(I,JMAX)=FS(I,JMAX-1);
ENDDO

RETURN
END

!------------------------------------------------------------

SUBROUTINE CALVEL
!     Purpose:Calculate tentative velocity field
!     at current time level
USE FIELD
IMPLICIT NONE
DOUBLE PRECISION ::XFAC,YFAC,FUX,FUY,FVX,FVY,VISU,VISV,DIFFE,DIFFN,DIFFS,DIFFW,DPX,DPY


XFAC = 1.0/(DELX*DELX)
YFAC = 1.0/(DELY*DELY)

DO I=2,IM1
  DO J=2,JM1

	!Convective terms    
    FUX=((U(I,J)+U(I+1,J))**2-(U(I-1,J)+U(I,J))**2)*0.25*RDX
    FUY=((V(I,J)+V(I+1,J))*(U(I,J)+U(I,J+1))-(V(I,J-1)+V(I+1,J-1))*(U(I,J-1)+U(I,J)))*0.25*RDY
    FVX=((U(I,J)+U(I,J+1))*(V(I,J)+V(I+1,J))-(U(I-1,J)+U(I-1,J+1))*(V(I-1,J)+V(I,J)))*0.25*RDX
    FVY=((V(I,J)+V(I,J+1))**2-(V(I,J-1)+V(I,J))**2)*0.25*RDY 
    
	!Viscous terms 
	! U
    DIFFN=U(I,J+1)-U(I,J)
    DIFFS=U(I,J)-U(I,J-1)
    DIFFE=U(I+1,J)-U(I,J)
    DIFFW=U(I,J)-U(I-1,J)

	! hardcode s-l BC
    IF ((FNP(I,J+1)+FNP(I+1,J+1)).NE.2) THEN !North border
	DIFFN=-2*U(I,J)
    ENDIF
    IF ((FNP(I,J-1)+FNP(I+1,J-1)).NE.2) THEN !South border
	DIFFS=2*U(I,J)
    ENDIF
    IF (FNP(I+1,J).EQ.0) THEN !East border
	DIFFE=-2*U(I,J)
    ENDIF
    IF (FNP(I,J).EQ.0) THEN !West border
	DIFFW=2*U(I,J)
    ENDIF
    VISU=(DIFFE-DIFFW)*XFAC + (DIFFN-DIFFS)*YFAC

	! V
    DIFFN=V(I,J+1)-V(I,J)
    DIFFS=V(I,J)-V(I,J-1)
    DIFFE=V(I+1,J)-V(I,J)
    DIFFW=V(I,J)-V(I-1,J)
	! hardcode s-l BC
    IF (FNP(I,J+1).EQ.0) THEN !North border
	DIFFN=-2*V(I,J)
    ENDIF
    IF (FNP(I,J).EQ.0) THEN !South border
	DIFFS=2*V(I,J)
    ENDIF
    IF ((FNP(I+1,J)+FNP(I+1,J+1)).NE.2) THEN !East border
	DIFFE=-2*V(I,J)
    ENDIF
    IF ((FNP(I-1,J)+FNP(I-1,J+1)).NE.2) THEN !West border
	DIFFW=2*V(I,J)
    ENDIF
    VISV=(DIFFE-DIFFW)*XFAC + (DIFFN-DIFFS)*YFAC

	! Bouynacy forces
    FB=0.5*RA*PR*(TEMP(I,J)+TEMP(I,J+1))


	! Pressure forces
    DPX=P(I+1,J)-P(I,J)
    DPY=P(I,J+1)-P(I,J)
	!hardcode s-l BC
    IF ((FNP(I,J)+FNP(I+1,J)).LT.2) THEN !East/West border
	DPX=0.0
    ENDIF
    IF ((FNP(I,J)+FNP(I,J+1)).LT.2) THEN !North/South border
	DPY=0.0
    ENDIF

	! Premliminary velocities
    IF ((FNP(I,J)+FNP(I+1,J)).NE.0) THEN
      U(I,J)=U(I,J) + DELT*(-DPX*RDX - FUX - FUY + PR*VISU)
    ELSE
      U(I,J)=0.0
    ENDIF

    IF ((FNP(I,J)+FNP(I,J+1)).NE.0) THEN
      V(I,J)=V(I,J) + DELT*(-DPY*RDY - FVX - FVY + PR*VISV + FB)
    ELSE
      V(I,J)=0.0
    ENDIF

  ENDDO   
ENDDO

   
RETURN
END

!-----------------------------------------------------------------

SUBROUTINE GEN_T_MAT 
!     purpose:
!     To construct the sparse coefficient matrix A_T for solving temperature
USE FIELD
IMPLICIT NONE
INTEGER :: N
DOUBLE PRECISION :: APO,AP,AN,AS,AE,AW


      
  N = 0

  DO I=2,IM1
    DO J=2,JM1
      APO = DELX*DELY/DELT
      AE = 0.5*U(I,J)*DELY - KON(I,J)*DELY/DELX - (KON(I+1,J)-KON(I-1,J))*DELY/(4*DELX)
      AW = -0.5*U(I-1,J)*DELY - KON(I,J)*DELY/DELX + (KON(I+1,J)-KON(I-1,J))*DELY/(4*DELX)
      AN = 0.5*V(I,J)*DELX - KON(I,J)*DELX/DELY - (KON(I,J+1)-KON(I,J-1))*DELX/(4*DELY)
      AS = -0.5*V(I,J-1)*DELX - KON(I,J)*DELX/DELY + (KON(I,J+1)-KON(I,J-1))*DELX/(4*DELY)
      AP = 2*KON(I,J)*(DELX/DELY + DELY/DELX)

      ! center node 
      N = N + 1
      TAVAL(N) = APO + AE + AW + AN + AS + 2*AP
      TI(N) = NNT(I,J)
      TJ(N) = NNT(I,J)

      !At Boundary walls, grad(T)=0
      !IF (I.EQ.(IM1)) THEN !East node
      !  TAVAL(N) = TAVAL(N) + AE
      !ENDIF
      !IF (I.EQ.2) THEN !West node
      !  TAVAL(N) = TAVAL(N) + AW
      !ENDIF


      ! Neighbour nodes in domain
      IF(I.LT.IM1) THEN !East Node
        N = N + 1
        TAVAL(N) = AE
        Ti(N) = NNT(I,J)
        Tj(N) = NNT(I+1,J)

      ENDIF
      IF(I.GT.2) THEN !West Node
        N = N + 1
        TAVAL(N) = AW
        Ti(N) = NNT(I,J)
        Tj(N) = NNT(I-1,J)

      ENDIF
      IF(J.LT.JM1) THEN !North Node
        N = N + 1
        TAVAL(N) = AN
        Ti(N) = NNT(I,J)
        Tj(N) = NNT(I,J+1)

      ENDIF
      IF(J.GT.2) THEN !South Node
        N = N + 1
        TAVAL(N) = AS
        Ti(N) = NNT(I,J)
        Tj(N) = NNT(I,J-1)

      ENDIF
      
    TEMP_VEC(NNT(I,J))=TEMP(I,J)




    ENDDO
  ENDDO    


RETURN
END

!------------------------------------------------------------
SUBROUTINE GEN_TB_VEC
!     purpose:
!     To construct b in Ax=b
USE FIELD
IMPLICIT NONE

DOUBLE PRECISION :: APO,AE,AW,AN,AS


TB(:)=0.0
APO = DELX*DELY/DELT


DO I=2,IM1
  DO J=2,JM1

    TB(NNT(I,J))= APO*TEMP_P(I,J)


    !At Boundary walls, T=known

    IF (I.EQ.IM1) THEN !East node
      AE = 0.5*U(I,J)*DELY - KON(I,J)*DELY/DELX - (KON(I+1,J)-KON(I-1,J))*DELY/(4*DELX)
      TB(NNT(I,J)) = TB(NNT(I,J)) - AE*TEMP_P(I+1,J)

    ENDIF
    IF (I.EQ.2) THEN !West node
      AW = -0.5*U(I-1,J)*DELY - KON(I,J)*DELY/DELX + (KON(I+1,J)-KON(I-1,J))*DELY/(4*DELX)
      TB(NNT(I,J)) = TB(NNT(I,J)) - AW*TEMP_P(I-1,J)
    ENDIF
    IF (J.EQ.JM1) THEN !North node
      AN = 0.5*V(I,J)*DELX - KON(I,J)*DELX/DELY - (KON(I,J+1)-KON(I,J-1))*DELX/(4*DELY)
      TB(NNT(I,J)) = TB(NNT(I,J)) - AN*TEMP_P(I,J+1)

    ENDIF
    IF (J.EQ.2) THEN !South node
      AS = -0.5*V(I,J-1)*DELX - KON(I,J)*DELX/DELY + (KON(I,J+1)-KON(I,J-1))*DELX/(4*DELY)
      TB(NNT(I,J)) = TB(NNT(I,J)) - AS*TEMP_P(I,J-1)
    ENDIF


  ENDDO
ENDDO


RETURN
END

!------------------------------------------------------------
SUBROUTINE CORR_FS
!     purpose:
!     Correct solide fraction
USE FIELD
USE SOLVEFIELD
IMPLICIT NONE

DOUBLE PRECISION :: VAL1, VAL2, F_INV_PURE

ITMAX_TT=30
EPSI_TT=1E-6

! initial guesses for Xi
DO I=2,IM1
  DO J=2,JM1
    XI(I,J) = FS(I,J) - FS_P(I,J) - STE*(F_INV_PURE(FS(I,J),TM) - TEMP(I,J))
    XI_N(I,J) = FS_N(I,J) - FS_P(I,J) - STE*(F_INV_PURE(FS_N(I,J), TM) - TEMP(I,J))
  ENDDO
ENDDO


! Secant method
DO ITER_TT=1,ITMAX_TT
  ! reset residual
  RES_TT=0.0

  DO I=2,IM1
    DO J=2,JM1

      ! Find phase change in fluid domain only
      IF (GEO(I,J).EQ.0) THEN
        ! Hold values for next iteration
        VAL1=FS(I,J)
        VAL2=XI(I,J)
      
        ! next value
        FS(I,J) = FS(I,J) - XI(I,J)*(FS(I,J)-FS_N(I,J))/(XI(I,J)-XI_N(I,J))
        XI(I,J) = FS(I,J) - FS_P(I,J) - STE*(F_INV_PURE(FS(I,J),TM) - TEMP(I,J))

        ! Replace old values
        FS_N(I,J) = VAL1
        XI_N(I,J) = VAL2

        ! residual
        RES_TT = RES_TT + ABS(XI(I,J)-XI_N(I,J))
      ENDIF

    ENDDO
  ENDDO

  ! Convergence test
  IF (RES_TT.LT.EPSI_TT) THEN
    EXIT
  ENDIF

ENDDO


! limit Solid Fraction
DO I=2,IM1
  DO J=2,JM1
    FS(I,J) = MIN( 1.0 , MAX( 0.0 , FS(I,J) ) )
  ENDDO
ENDDO



RETURN
END


!------------------------------------------------------------
SUBROUTINE TITER
!     Purpose: Calculate temperature field
!     at current time level
USE FIELD
USE SOLVEFIELD
IMPLICIT NONE


!Store previous time step temperature field
TEMP_P=TEMP
FS_P=FS

! second starting guess of secant method
FS_N=FS+0.1


!Generate system matrix
CALL GEN_T_MAT
  
!generate b-Vector
CALL GEN_TB_VEC


!Calculate tentative Temperature field
CALL CGSOLVER(TAVAL,TI,TJ,TB,TEMP_VEC,UCOUNT_T,ENTRIESA_T,ITER_T,RES_T)

!Update tentative temperature field
DO I=2,IM1
  DO J=2,JM1
    TEMP(I,J) = TEMP_VEC(NNT(I,J))
  ENDDO
ENDDO

!Solid fraction Correction
CALL CORR_FS

!Correct Temperature field
DO I=2,IM1
  DO J=2,JM1
    TEMP(I,J) = TEMP(I,J) + (FS(I,J) - FS_P(I,J))/STE
  ENDDO
ENDDO

! Impose BC
CALL BCTEMP

 



  
RETURN
END

!------------------------------------------------------------
SUBROUTINE GEN_A_MAT(UNKNOWNS) 
!     purpose:
!     To construct the sparse coefficient matrix A
USE FIELD
IMPLICIT NONE

INTEGER :: UNKNOWNS,N
DOUBLE PRECISION :: AP,AN,AS,AE,AW


! Find entries in A so that A can be allocated
  N = 0

  DO I=2,IM1
    DO J=2,JM1
      !current cell is a fluid cell
      IF (FNP(I,J).EQ.1) THEN
       ! center
       N = N + 1


	!fluid cell to the south
        IF(FNP(I,J-1).EQ.1) THEN
	  ! Add South node
	  N = N + 1
	ENDIF

	!fluid cell to the north
        IF(FNP(I,J+1).EQ.1) THEN
	  ! Add North node
	  N = N + 1
    	ENDIF

	!fluid cell to the west
        IF(FNP(I-1,J).EQ.1) THEN
	  ! Add West node
	  N = N + 1
    	ENDIF

	!fluid cell to the east
        IF(FNP(I+1,J).EQ.1) THEN
	  ! Add East node
	  N = N + 1
    	ENDIF        
      ENDIF
    ENDDO
  ENDDO
  

  ENTRIESA = N

  ALLOCATE(AVAL(ENTRIESA),AI(ENTRIESA),AJ(ENTRIESA))
  ALLOCATE(B(UNKNOWNS),DELP(UNKNOWNS))



  DELP(:) = 0.0
  
  N = 0

  AE = 1/(DELX**2)
  AW = 1/(DELX**2)
  AN = 1/(DELY**2)
  AS = 1/(DELY**2)
  AP = -(AE+AW+AN+AS)

  DO I=2,IM1
    DO J=2,JM1
      !Current node is a fluid node
      IF (FNP(I,J).EQ.1) THEN
        !center node 
        N = N + 1
        AVAL(N) = AP
        AI(N) = NN(I,J)
        AJ(N) = NN(I,J)

	!At Boundary wall/s-l interphase, grad(P)=0
        IF (FNP(I+1,J).EQ.0) THEN !East node
          Aval(N) = Aval(N) + AE
        ENDIF
        IF (FNP(I-1,J).EQ.0) THEN !West node
          Aval(N) = Aval(N) + AW
        ENDIF
        IF (FNP(I,J+1).EQ.0) THEN !North node
          Aval(N) = Aval(N) + AN
        ENDIF
        IF (FNP(I,J-1).EQ.0) THEN !South node
          Aval(N) = Aval(N) + AS
        ENDIF


	! Neighbour nodes are fluid nodes
        IF(FNP(I+1,J).EQ.1) THEN !East Node
          N = N + 1
          Aval(N) = AE
          Ai(N) = NN(I,J)
          Aj(N) = NN(I+1,J)

        ENDIF
        IF(FNP(I-1,J).EQ.1) THEN !West Node
          N = N + 1
          Aval(N) = AW
          Ai(N) = NN(I,J)
          Aj(N) = NN(I-1,J)

        ENDIF
        IF(FNP(I,J+1).EQ.1) THEN !North Node
          N = N + 1
          Aval(N) = AN
          Ai(N) = NN(I,J)
          Aj(N) = NN(I,J+1)

        ENDIF
        IF(FNP(I,J-1).EQ.1) THEN !South Node
          N = N + 1
          Aval(N) = AS
          Ai(N) = NN(I,J)
          Aj(N) = NN(I,J-1)

        ENDIF
      ENDIF

    ENDDO
  ENDDO    


PREFNODE = 1  ! Setting the node at which the pressure will always be zero

IF (PREFNODE.NE.0) THEN
  DO N=1,ENTRIESA		! Setting reference pressure to zero
    IF((Ai(N).EQ.PREFNODE).AND.(Aj(N).EQ.PREFNODE)) THEN
      Aval(N) = -1E-30
    ELSEIF((Ai(N).EQ.PREFNODE).OR.(Aj(N).EQ.PREFNODE)) THEN
      Aval(N) = 0
    ENDIF
  ENDDO     
ENDIF

RETURN
END

!--------------------------------------------------------------------------

SUBROUTINE GEN_B_VEC
!     purpose:
!     To construct b in Ax=b
USE FIELD
IMPLICIT NONE

DOUBLE PRECISION :: DIV


DO I=2,IM1
  DO J=2,JM1

    ! Current node is a fluid node
    IF (FNP(I,J).EQ.1) THEN

      DIV = RDX*(U(I,J)-U(I-1,J)) + RDY*(V(I,J)-V(I,J-1))
      B(NN(I,J))=DIV/(DELT)
    ENDIF

  ENDDO
ENDDO



B(PREFNODE) = 0.0	! Assigning reference pressure 0 to pressure reference cell

RETURN
END

!------------------------------------------------------------
SUBROUTINE PITER
!     Purpose:To solve the Poissons equation of pressure
!     and adjust velocity and pressure fields so that 
!     continuity is maintained for every time step.
USE FIELD
USE SOLVEFIELD
IMPLICIT NONE


!Calculate tentative velocity field
CALL CALVEL

! impose BC
CALL BCVEL
CALL BCP  


! main Momentum loop (not running if there are no fluid nodes)
IF (UCOUNT.GT.0) THEN

  !Generate A matrix in Ax=b
  CALL GEN_A_MAT(UCOUNT)	

  !Generate b vector in Ax=b
  CALL GEN_B_VEC			

  !Solve system
  CALL CGSOLVER(AVAL,AI,AJ,B,DELP,UCOUNT,ENTRIESA,ITER_P,RES_P) 


  DO I=2,IM1-1	! Updating U-components, if fluid node
    DO J=2,JM1
      IF((FNP(I+1,J).EQ.1).AND.(FNP(I,J).EQ.1)) THEN
        U(I,J)=U(I,J) + DELT*(DELP(NN(I,J))-DELP(NN(I+1,J)) )/DELX
      ENDIF
    ENDDO
  ENDDO

  DO I=2,IM1		! Updating V-components, if fluid node
    DO J=2,JM1-1
      IF((FNP(I,J+1).EQ.1).AND.(FNP(I,J).EQ.1)) THEN
        V(I,J)=V(I,J) + DELT*( DELP(NN(I,J))-DELP(NN(I,J+1)) )/DELY
      ENDIF
    ENDDO
  ENDDO   


  DO I=2,IM1		! Updating pressure field, if fluid node
    DO J=2,JM1
      IF (FNP(I,J).EQ.1) THEN
        P(I,J) = P(I,J) + DELP(NN(I,J))
      ENDIF
    ENDDO
  ENDDO   

  ! impose BC
  CALL BCVEL
  CALL BCP

  ! Deallocate (size is changing)



  DEALLOCATE(DELP)

  DEALLOCATE(AVAL)

  DEALLOCATE(B)

  DEALLOCATE(AJ)

  DEALLOCATE(AI)


ENDIF
RETURN
END 

!------------------------------------------------------------------------------------






! ========= CONJUGATE GRADIENT SOLVER =========


SUBROUTINE CGSOLVER(Aval,Ai,Aj,b,x,unknowns,entriesA, iter, res)	! Solves the system Ax=b with the conjugate gradients method

  IMPLICIT NONE
  INTEGER :: unknowns, entriesA,itmax,iter
  DOUBLE PRECISION, DIMENSION (entriesA) :: Aval
  INTEGER, DIMENSION (entriesA) :: Ai, Aj
  DOUBLE PRECISION, DIMENSION (unknowns) :: tempvec, x, r, d, b
                               
  DOUBLE PRECISION :: alpha, beta, tempscal1, tempscal2,epsi,res
  

  ! iteration parameters
  epsi=1E-7
  itmax=unknowns
  iter=0

  x(:) = 0.0
  r(:) = b(:)
  d(:) = r(:) 
  res = sum(abs(r))
  IF(res.LT.epsi) THEN		! Convergence test
     RETURN
  ENDIF


  DO iter=1,itmax
    CALL VECVEC(r,r,tempscal1,unknowns)			! (r'*r)
    CALL MATVEC(Aval,Ai,Aj,d,tempvec,unknowns,entriesA)	! (A*d)
    CALL VECVEC(d,tempvec,tempscal2,unknowns)		! (d'*A*d)
    alpha = tempscal1/tempscal2				! alpha = (r'*r)/(d'*A*d)
    x(:) = x(:) + alpha*d(:)				! x = x + alpha*d
    tempvec(:) = r(:) - alpha*tempvec(:)   		! r_new = r - alpha*A*d
    CALL VECVEC(tempvec,tempvec,tempscal2,unknowns)	! (r_new'*r_new)
    beta = tempscal2/tempscal1				! beta = (r_new'*r_new)/(r'*r)
    r(:) = tempvec(:)					! r = r_new
    d(:) = r(:) + beta*d(:)  				! d = r + beta*d

    res = sum(abs(r))
    IF(res.LT.epsi) THEN		! Convergence test
      EXIT
    ENDIF

  ENDDO
  

RETURN
END





SUBROUTINE MATVEC(Aval,Ai,Aj,u,x,m,n)	! Matrix-vector multiplication, x=Au  -- A: matrix, u: vector, x: return vector, m: length of u, n: entries in A
  IMPLICIT NONE
  INTEGER :: i, m, n
  DOUBLE PRECISION, DIMENSION(n) :: Aval 
  INTEGER, DIMENSION(n) :: Ai, Aj 
  DOUBLE PRECISION, DIMENSION(m) :: u,x
  
  x(:) = 0
  DO i=1,n
    x(Ai(i)) = x(Ai(i)) + Aval(i)*u(Aj(i))
  ENDDO
  
RETURN
END



SUBROUTINE VECVEC(u,v,x,m)	! Vector-vector multiplication, x=uv -- u: vector1, v: vector2, x: return scalar, m: size of u and v
  IMPLICIT NONE
  INTEGER :: i, m
  DOUBLE PRECISION, DIMENSION(m) :: u, v
  DOUBLE PRECISION :: x
  
  x = 0.0
  DO i=1,m
    x = x + u(i)*v(i)
  ENDDO
  
RETURN
END



!------------------------------------------------------------------


FUNCTION F_INV_PURE(x,ym)			
! inverse of solid fraction temperature relation for pure substance		
! x = Solid fraction
! y = Melting temperature
  DOUBLE PRECISION :: F_INV_PURE,x,ym
  

  F_INV_PURE = ym

RETURN
END FUNCTION F_INV_PURE








