
!Physical variables
MODULE FIELD
IMPLICIT NONE
SAVE

INTEGER,PARAMETER::NINP=21,ANIMATION=22, RUNDATA=23, INTERP=24
INTEGER::IBAR,JBAR,KBAR,IM1,JM1,KM1,IMAX,JMAX,KMAX,I,J,K,NFRAME,NTHFRAME,&
UCOUNT,UCOUNT_T,ENTRIESA,FLAG,PREFNODE,ENTRIESA_T
DOUBLE PRECISION :: DELT,T,TMAX,DELX,DELY,DELZ,RELAX,XMAX,YMAX,ZMAX,RDX,RDY,RDZ,&
TH,TC,TM,ADVX,ADVY,DIFFX,DIFFY,PR,RA,STE,VF,MC,STC,FB,TH_RE,TM_RE,TC_RE,T_MID,H,&
TAU,FSLIM,RHO_GAL1,RHO_GAL2,CAP_GAL1,CAP_GAL2,KL_GAL1,KL_GAL2,CPU_START,CPU_STOP,DELTN
DOUBLE PRECISION,ALLOCATABLE,DIMENSION (:,:,:) :: U,V,W,P,TEMP,TEMP_P,FS,&
FS_P,FS_N,XI,XI_N,KON
INTEGER, ALLOCATABLE, DIMENSION (:,:,:) :: NN,NNT,FNP
DOUBLE PRECISION, ALLOCATABLE, DIMENSION (:) :: B,DELP,AVAL,PX,PY,TB,TAVAL,TEMP_VEC,X,Y,Z
INTEGER, ALLOCATABLE, DIMENSION (:) :: AI,AJ,TI,TJ

END MODULE FIELD

!Pressure solver variables
MODULE SOLVEFIELD
IMPLICIT NONE
SAVE

INTEGER :: ITER_P,ITER_T,ITER_TT,ITMAX_TT,ITMAX_T,ITMAX_P
DOUBLE PRECISION :: RES_P,RES_T,RES_TT,EPSI_TT,EPSI_T,EPSI_P

END MODULE SOLVEFIELD



!------------------------------------------------------------------

FUNCTION F_INV_PURE(x,y)			
! Inverse of solid fraction temperature relation for substance
! Here: pure material		
! x = Solid fraction
! y = Melting temperature
  DOUBLE PRECISION :: F_INV_PURE,x,y
  

  F_INV_PURE = y

RETURN
END FUNCTION F_INV_PURE

!-------------------------------------------------------------------












PROGRAM CAVITY3D
!-------------------------------------------------------------
!     SOLVES THE NAVIER-STOKES EQUATIONS ON A 3-DIMENSIONAL
!     RECTANGULAR GRID WITH A MELTING FLUID
!     Version March 2018 Olav Rømcke
!------------------------------------------------------------
	



USE FIELD
USE SOLVEFIELD

IMPLICIT NONE


CALL INPUT                ! Read input
CALL PREPARE_TECPLOT_FILE ! Prepare files for writing data
CALL INITVPT  		  ! Define initial Thermal and momentum fields.
CALL BCVEL    		  ! Impose boundary conditions by calling BCVEL.
CALL BCTEMP   		  ! Impose boundary conditions by calling BCTEMP.
CALL NODE_NUMBERS_TEMP	  ! define node numbers for temperature field

WRITE(*,*)'Simulation Input:'
WRITE(*,*)'-------------------------------------------------------------------------------------'
WRITE(*,'(3(A,G12.5))')' Pr         = ',PR,      ' |  Ra          = ',RA,      ' |  Ste         = ',STE
WRITE(*,'(3(A,G12.5))')' H          = ',H,       ' |  End Time    = ',TMAX,    ' |  Tau         = ',TAU
WRITE(*,'(3(A,G12.5))')' T Cold     = ',TC_RE,   ' |  T Hot       = ',TH_RE,   ' |  T Ref       = ',TM_RE
WRITE(*,'(3(A,G12.5))')' C Liq      = ',CAP_GAL1,' |  C Sol       = ',CAP_GAL2,' |  R Liq       = ',RHO_GAL1
WRITE(*,'(3(A,G12.5))')' R Sol      = ',RHO_GAL2,' |  K Liq       = ',KL_GAL1, ' |  K Sol       = ',KL_GAL2
WRITE(*,'(3(A,G12.5))')' Epsilon P  = ',EPSI_P,  ' |  Epsilon T   = ',EPSI_T,  ' |  Epsilon TT  = ',EPSI_TT
WRITE(*,'(3(A,I12))')  ' Max Iter P = ',ITMAX_P, ' |  Max Iter T  = ',ITMAX_T, ' |  Max Iter TT = ',ITMAX_TT
WRITE(*,*)'-------------------------------------------------------------------------------------'
WRITE(*,*)




!     ==== START OF GLOBAL LOOP ====

! Start Stopwatch
CALL CPU_TIME(CPU_START)

  
DO WHILE(.TRUE.)

!----Recording values----
  IF(MOD(NFRAME,NTHFRAME).EQ.0) THEN
    CALL WRITE_TO_TECPLOT 			! Animation
  ENDIF
  CALL WRITE_TO_TECPLOT2  			! Running (meta) data
  
!-----Next time step----
  NFRAME = NFRAME + 1
  T=T+DELT
  CALL DEF_NODES			  	! Defining nodes
  CALL PHYS_VAL					! Physical values
  CALL PITER					! Momentum calculations
  CALL TITER					! Energy calculations
  CALL GLOB_VAL					! Global values

!----Write to screen----
  WRITE(*,*)
  WRITE(*,*)
  WRITE(*,'(A,G12.5)') ' Time = ',T
  WRITE(*,*)'-------------------------------------------------------------------------------------'
  WRITE(*,'(3(A,I12))')  ' PC-iter = ',ITER_P,' |  TT-iter = ',ITER_TT,' |  T-iter = ',ITER_T
  WRITE(*,'(3(A,G12.5))')' PC-res  = ',RES_P, ' |  TT-res  = ',RES_TT, ' |  T-res  = ',RES_T
  WRITE(*,*)'-------------------------------------------------------------------------------------'

  IF(T.GE.TMAX)EXIT
END DO


!     ==== END OF GLOBAL LOOP ====

! End Stopwatch
CALL CPU_TIME(CPU_STOP)

! Write Experimental data
CALL WRITE_COMP_DAT

! end routine
WRITE(*,*)
WRITE(*,*)
WRITE(*,*)
WRITE(*,*)'SIMULATION FINISHED!'
WRITE(*,*)
WRITE(*,*)'Simulation Input:'
WRITE(*,*)'-------------------------------------------------------------------------------------'
WRITE(*,'(3(A,G12.5))')' Pr         = ',PR,      ' |  Ra          = ',RA,      ' |  Ste         = ',STE
WRITE(*,'(3(A,G12.5))')' H          = ',H,       ' |  End Time    = ',TMAX,    ' |  Tau         = ',TAU
WRITE(*,'(3(A,G12.5))')' T Cold     = ',TC_RE,   ' |  T Hot       = ',TH_RE,   ' |  T Ref       = ',TM_RE
WRITE(*,'(3(A,G12.5))')' C Liq      = ',CAP_GAL1,' |  C Sol       = ',CAP_GAL2,' |  R Liq       = ',RHO_GAL1
WRITE(*,'(3(A,G12.5))')' R Sol      = ',RHO_GAL2,' |  K Liq       = ',KL_GAL1, ' |  K Sol       = ',KL_GAL2
WRITE(*,'(3(A,G12.5))')' Epsilon P  = ',EPSI_P,  ' |  Epsilon T   = ',EPSI_T,  ' |  Epsilon TT  = ',EPSI_TT
WRITE(*,'(3(A,I12))')  ' Max Iter P = ',ITMAX_P, ' |  Max Iter T  = ',ITMAX_T, ' |  Max Iter TT = ',ITMAX_TT
WRITE(*,*)'-------------------------------------------------------------------------------------'
WRITE(*,*)
WRITE(*,*)'Global End Results:'
WRITE(*,*)'-------------------------------------------------------------------------------------'
WRITE(*,'(A,G12.5,A)')'Duration:                 ', CPU_STOP-CPU_START, ' sec '
WRITE(*,'(A,G12.5,A)')'Global Mass Conservation: ', MC, '  - '
WRITE(*,'(A,G12.5,A)')'Global Solid Fraction:    ', VF, '  - '
WRITE(*,*)'-------------------------------------------------------------------------------------'

DEALLOCATE (U,V,W,P,TEMP,TEMP_P,X,Y,Z,FS,FS_P,NN,FNP,&
NNT,TAVAL,TI,TJ,TB,TEMP_VEC,KON,PX,PY,XI,XI_N, FS_N)

STOP
END PROGRAM


!------------------------------------------------------------

SUBROUTINE INPUT
!     purpose:
!     To read input data and compute parameters from input file 
USE FIELD
USE SOLVEFIELD
IMPLICIT NONE
INTEGER ::IFAIL
DOUBLE PRECISION ::DUMMY1, DUMMY2, DUMMY3


!Read input file
OPEN (UNIT=NINP,FILE='SIMINPUT3D.INP',FORM='FORMATTED')
READ(NINP,1000)
READ(NINP,1020)TMAX,DELT,DELTN,DELX,DELY,DELZ,XMAX,YMAX,ZMAX,RA,PR,STE,TH_RE,&
TM_RE,TC_RE,KL_GAL1,KL_GAL2,CAP_GAL1,CAP_GAL2,RHO_GAL1,RHO_GAL2,H,&
EPSI_P,EPSI_T,EPSI_TT,DUMMY1,DUMMY2,DUMMY3


!Define grid
IBAR=NINT(DELX)
JBAR=NINT(DELY)
KBAR=NINT(DELZ)
DELX=XMAX/DELX
DELY=YMAX/DELY
DELZ=ZMAX/DELZ
IM1=IBAR+1
JM1=JBAR+1
KM1=KBAR+1
IMAX=IBAR+2
JMAX=JBAR+2
KMAX=KBAR+2


!Iteration parameters
ITMAX_P = NINT(DUMMY1)
ITMAX_T = NINT(DUMMY2)
ITMAX_TT = NINT(DUMMY3)

!Recording parameter, When to capture physics
NTHFRAME = CEILING(DELTN/DELT)
NFRAME = 0


!   Initiating parameters and fields

RDX  = 1.0/DELX
RDY  = 1.0/DELY
RDZ  = 1.0/DELZ

!set time to zero initially
T=0.0

! Time constant	  [s]
TAU = (H**2*RHO_GAL1*CAP_GAL1/KL_GAL1)

!set mass conservation to zero initially
MC=0

! solid liquid interphase limit
FSLIM=0.5



! Temperature scale
TH=(TH_RE-TC_RE)/(TH_RE-TC_RE)
TC=(TC_RE-TC_RE)/(TH_RE-TC_RE)
TM=(TM_RE-TC_RE)/(TH_RE-TC_RE)


ALLOCATE (U(IBAR+2,JBAR+2,KBAR+2),V(IBAR+2,JBAR+2,KBAR+2),W(IBAR+2,JBAR+2,KBAR+2),&
P(IBAR+2,JBAR+2,KBAR+2),TEMP(IBAR+2,JBAR+2,KBAR+2),TEMP_P(IBAR+2,JBAR+2,KBAR+2),&
KON(IBAR+2,JBAR+2,KBAR+2),X(IBAR+2),Y(JBAR+2),Z(KBAR+2),FS(IBAR+2,JBAR+2,KBAR+2),&
FS_P(IBAR+2,JBAR+2,KBAR+2),FS_N(IBAR+2,JBAR+2,KBAR+2),XI(IBAR+2,JBAR+2,KBAR+2),&
XI_N(IBAR+2,JBAR+2,KBAR+2),NN(IBAR+2,JBAR+2,KBAR+2),FNP(IBAR+2,JBAR+2,KBAR+2),&
NNT(IBAR+2,JBAR+2,KBAR+2),PX(JBAR),PY(JBAR))

1000 FORMAT(//////)
1020 FORMAT(10(43X,F20.0/))
RETURN
END 
!------------------------------------------------------------
SUBROUTINE PREPARE_TECPLOT_FILE
USE FIELD
IMPLICIT NONE

! coordinate vectors
DO I = 1,IMAX
  DO J = 1,JMAX
    DO K = 1,KMAX
      X(I) = (I-1.5)*DELX
      Y(J) = (J-1.5)*DELY
      Z(K) = (K-1.5)*DELZ
    ENDDO
  ENDDO
ENDDO

OPEN (UNIT=ANIMATION,FILE='TECPLOT.DAT',FORM='FORMATTED')  ! Output file
WRITE(ANIMATION,'(A)')'VARIABLES = "X","Y","Z","U","V","W","P","T","FS","K","NN","NNT","FP"'
WRITE(ANIMATION,*)''

OPEN (UNIT=RUNDATA,FILE='TECPLOT2.DAT',FORM='FORMATTED')   ! Output file
WRITE(RUNDATA,*)'TITLE     = "Run Data"'
WRITE(RUNDATA,*)'VARIABLES = "T", "MC", "VF", "RES_P", "ITER_P", "RES_T", "ITER_T", "RES_TT", "ITER_TT"'
WRITE(RUNDATA,*)''

OPEN (UNIT=INTERP,FILE='TECPLOT3.DAT',FORM='FORMATTED')    ! Output file
WRITE(INTERP,*)'TITLE     = "Inter Phase"'
WRITE(INTERP,*)'VARIABLES = "X", "Y"'
WRITE(INTERP,*)''

RETURN
END 

!------------------------------------------------------------
SUBROUTINE WRITE_TO_TECPLOT
!     purpose:
!     To create a dat file for use in tecplot
!     Will be used to create an animation
USE FIELD
IMPLICIT NONE
INTEGER ::K_COORD

K_COORD=CEILING(KMAX/2.0)

! Calculate solid-liquid Phase front x-y coordinates at mid z plane
PX(:)=0.0
PY(:)=0.0

DO J=2,JM1
  PY(J-1)=Y(J)
  DO I=2,IM1
     IF ((FS(I,J,K_COORD).GT.(FSLIM)).AND.(FS(I+1,J,K_COORD).LE.(FSLIM))) THEN
	PX(J-1)=X(I) + DELX*(FSLIM-FS(I,J,K_COORD))/(FS(I+1,J,K_COORD)-FS(I,J,K_COORD))
     ENDIF
     IF ((FS(I,J,K_COORD).LE.(FSLIM)).AND.(FS(I+1,J,K_COORD).GT.(FSLIM))) THEN
	PX(J-1)=X(I) - DELX*(FS(I+1,J,K_COORD)-FSLIM)/(FS(I+1,J,K_COORD)-FS(I,J,K_COORD))
     ENDIF
     
  ENDDO
ENDDO


! Write to INTERPHASE file
WRITE(INTERP,*)'ZONE ','T="time = ',T*TAU,'" J=',(JM1-1),', DATAPACKING=POINT'
DO J=1,JM1-1
    WRITE(INTERP,'(F10.7, A, F10.7)')PX(J),' ',PY(J)
ENDDO
WRITE(INTERP,*)


! Write to ANIMATION file
WRITE(ANIMATION,'(A, F10.7, 3(A, I4), A, F10.7)')'ZONE T="time = ',T,&
'" I=',(IM1-1),', J=',(JM1-1),', K=',(KM1-1),', DATAPACKING=POINT, SOLUTIONTIME=',T

WRITE(ANIMATION,*)
WRITE(ANIMATION,*)
WRITE(ANIMATION,*)


DO K=2,KM1
  DO J=2,JM1
    DO I=2,IM1 

      WRITE(ANIMATION,'(3(F7.5, A), 7(G14.5, A), 2(I6,A), I6)')&
      X(I),' ',Y(J),' ',Z(K),' ',((U(I-1,J,K)+U(I,J,K))/2),' ',&
      ((V(I,J-1,K)+V(I,J,K))/2),' ',((W(I,J,K-1)+W(I,J,K))/2),' ',P(I,J,K),' ',&
      TEMP(I,J,K),' ',FS(I,J,K),' ',KON(I,J,K),' ',NN(I,J,K),' ',NNT(I,J,K),' ',&
      FNP(I,J,K)

    ENDDO
  ENDDO
ENDDO

WRITE(ANIMATION,*)
WRITE(ANIMATION,*)
WRITE(ANIMATION,*)



RETURN
END 

!------------------------------------------------------------
SUBROUTINE WRITE_TO_TECPLOT2
!     purpose:
!     To create a dat file for use in tecplot
!     Will be used to create an animation
USE FIELD
USE SOLVEFIELD
IMPLICIT NONE


! Write to run data file
WRITE(RUNDATA,*)T,' ',MC,' ',VF,' ',RES_P,' ',ITER_P,' ',RES_T,' ',ITER_T,' ',RES_TT,' ',ITER_TT



RETURN
END 

!------------------------------------------------------------

SUBROUTINE INITVPT
!     Purpose:
!     To initiate the Thermal and Momentum fields 
USE FIELD
IMPLICIT NONE

  U(:,:,:)=0.0
  V(:,:,:)=0.0
  W(:,:,:)=0.0
  P(:,:,:)=0.0
  TEMP(:,:,:)=TC	! Temperature, not to be confused with temporary
  FS(:,:,:)=1.0         ! Must be consistent with initial temperature
  KON(:,:,:)=1.0 + (KL_GAL2-KL_GAL1)/KL_GAL1*FS(:,:,:)


RETURN
END

!------------------------------------------------------------

SUBROUTINE NODE_NUMBERS_TEMP
!     purpose:
!     To assign node numbers to thermal fields
!     UCOUNT_T = number of unknowns
!     NNT(i,j) = node number associated with node at position (i,j,k)
!     N = Number of entries in temperature matrix
USE FIELD

IMPLICIT NONE

INTEGER :: N

NNT(:,:,:) = 0
UCOUNT_T = 0		
N = 0
 
DO K=2,KM1
  DO J=2,JM1
    DO I=2,IM1

      UCOUNT_T = UCOUNT_T + 1
      NNT(I,J,K) = UCOUNT_T       

      ! center
      N = N + 1

      !not south wall
      IF(J.GT.2) THEN
      ! Add South node
        N = N + 1
      ENDIF

      !not north wall
      IF(J.LT.JM1) THEN
	! Add North node
	N = N + 1
      ENDIF

      !not west wall
      IF(I.GT.2) THEN
	! Add West node
	N = N + 1
      ENDIF

      !not east wall
      IF(I.LT.IM1) THEN
	! Add East node
	N = N + 1
      ENDIF

      !not back wall
      IF(K.GT.2) THEN
	! Add Back node
	N = N + 1
      ENDIF

      !Not front
      IF(K.LT.KM1) THEN
	! Add Front node
	N = N + 1
      ENDIF  
      
    ENDDO
  ENDDO
ENDDO

!Allocate system matrix and vectors
ENTRIESA_T = N
ALLOCATE(TAVAL(ENTRIESA_T),TI(ENTRIESA_T),TJ(ENTRIESA_T))
ALLOCATE(TB(UCOUNT_T),TEMP_VEC(UCOUNT_T))


RETURN
END


!------------------------------------------------------------

SUBROUTINE DEF_NODES
!     purpose:
!     To assign node type, Solid node or fluid node
USE FIELD
IMPLICIT NONE

FNP(:,:,:)=0
NN(:,:,:) = 0
UCOUNT = 0		


!Fluid nodes
DO I=2,IM1
  DO J=2,JM1
    DO K=2,KM1
      !Pressure cell
      IF (FS(I,J,K).LT.FSLIM) THEN
        FNP(I,J,K)=1
      ELSE
        FNP(I,J,K)=0
      ENDIF
      IF (FNP(I,J,K).EQ.1) THEN 
        UCOUNT = UCOUNT + 1
        NN(I,J,K) = UCOUNT
      ENDIF

    ENDDO
  ENDDO
ENDDO



RETURN
END
!------------------------------------------------------------

SUBROUTINE GLOB_VAL
!     purpose:
!     To find the area
USE FIELD
IMPLICIT NONE

VF=0.0

DO I=2,IM1
  DO J=2,JM1
    DO K=2,KM1

      ! volume fraction
      VF = VF + FS(I,J,K)*DELX*DELY*DELZ/(XMAX*YMAX*ZMAX)

      ! mass conservation
      MC = MC& 
         + DELT*((U(I,J,K)-U(I-1,J,K))*(DELY*DELZ)&
         +(V(I,J,K)-V(I,J-1,K))*(DELX*DELZ)&
         +(W(I,J,K)-W(I,J,K-1))*(DELX*DELY))&
         /(XMAX*YMAX*ZMAX)
    ENDDO
  ENDDO
ENDDO



RETURN
END
!------------------------------------------------------------

SUBROUTINE BCVEL
!     purpose:
!     To give boundary conditions to the velocities around the domain
USE FIELD
IMPLICIT NONE
DO I=1,IM1
  DO J=1,JM1
    DO K=1,KM1
    
      IF((FNP(I,J,K).EQ.0).OR.(FNP(I+1,J,K).EQ.0)) THEN
        U(I,J,K) = 0.0
      ENDIF

      IF((FNP(I,J,K).EQ.0).OR.(FNP(I,J+1,K).EQ.0)) THEN
        V(I,J,K) = 0.0
      ENDIF

      IF((FNP(I,J,K).EQ.0).OR.(FNP(I,J,K+1).EQ.0)) THEN
        W(I,J,K) = 0.0
      ENDIF

      !At Boundary walls no slip, no pen
      IF (I.EQ.(IM1)) THEN !East wall
        U(I,J,K) = 0.0
        V(I+1,J,K) = -V(I,J,K)
        W(I+1,J,K) = -W(I,J,K)
      ENDIF
      IF (I.EQ.2) THEN !South wall
        U(I-1,J,K) = 0.0
        V(I-1,J,K) = -V(I,J,K)
        W(I-1,J,K) = -W(I,J,K)
      ENDIF
      IF (J.EQ.(JM1)) THEN !North wall
        U(I,J+1,K) = -U(I,J,K) 
        V(I,J,K) = 0.0 
        W(I,J+1,K) = -W(I,J,K)
      ENDIF
      IF (J.EQ.2) THEN !South wall
        U(I,J-1,K) = -U(I,J,K)
        V(I,J-1,K) = 0.0
        W(I,J-1,K) = -W(I,J,K)
      ENDIF
      IF (K.EQ.(KM1)) THEN !Front wall
        U(I,J,K+1) = -U(I,J,K)
        V(I,J,K+1) = -V(I,J,K)
        W(I,J,K) = 0.0
      ENDIF
      IF (K.EQ.2) THEN !Back wall
        U(I,J,K-1) = -U(I,J,K)
        V(I,J,K-1) = -V(I,J,K)
        W(I,J,K-1) = 0.0
      ENDIF


    ENDDO
  ENDDO
ENDDO
RETURN
END
!------------------------------------------------------------

SUBROUTINE BCTEMP
!     purpose:
!     To give boundary conditions to the temp and fs
USE FIELD
IMPLICIT NONE


DO J=1,JMAX               !Impose boundary condition on X-planes
  DO K=1,KMAX
    !Left
    TEMP(1,J,K) = 2*TH - TEMP(2,J,K)
    FS(1,J,K) = FS(2,J,K)
    !Right
    TEMP(IMAX,J,K) = 2*TC - TEMP(IMAX-1,J,K)
    FS(IMAX,J,K)=FS(IM1,J,K)
  ENDDO
ENDDO 

DO I=1,IMAX               !Impose boudary condition on Y-planes
  DO K=1,KMAX
    !Bottom
    TEMP(I,1,K)=TEMP(I,2,K)
    FS(I,1,K)=FS(I,2,K)
    !Top
    TEMP(I,JMAX,K)=TEMP(I,JM1,K)
    FS(I,JMAX,K)=FS(I,JM1,K)
  ENDDO
ENDDO  

DO I=1,IMAX               !Impose boudary condition on Z-planes
  DO J=1,JMAX
    !Back
    TEMP(I,J,1)=TEMP(I,J,2)
    FS(I,J,1)=FS(I,J,2)
    !Front
    TEMP(I,J,KMAX)=TEMP(I,J,KM1)
    FS(I,J,KMAX)=FS(I,J,KM1)
  ENDDO
ENDDO   




RETURN
END

!------------------------------------------------------------

SUBROUTINE PHYS_VAL

USE FIELD
IMPLICIT NONE

DO I=1,IMAX
  DO J=1,JMAX
    DO K=1,KMAX
      ! Thermal conduction
      KON(I,J,K) = 1.0 + (KL_GAL2-KL_GAL1)/KL_GAL1*FS(I,J,K)
    ENDDO
  ENDDO
ENDDO

RETURN
END

!------------------------------------------------------------

SUBROUTINE VEL
!     Purpose:Calculate tentative velocity field
!     at current time level
USE FIELD
IMPLICIT NONE
DOUBLE PRECISION ::XFAC,YFAC,ZFAC,FUX,FUY,FUZ,FVX,FVY,FVZ,FWX,FWY,FWZ,&
VISU,VISV,VISW,DIFFE,DIFFW,DIFFN,DIFFS,DIFFF,DIFFB,DPX,DPY,DPZ

DOUBLE PRECISION, DIMENSION(IMAX,JMAX,KMAX) :: DELTAU, DELTAV, DELTAW

XFAC = 1.0/(DELX*DELX)
YFAC = 1.0/(DELY*DELY)
ZFAC = 1.0/(DELZ*DELZ)

DELTAU(:,:,:) = 0.0
DELTAV(:,:,:) = 0.0
DELTAW(:,:,:) = 0.0

DO I=2,IM1
  DO J=2,JM1
    DO K=2,KM1



    ! -------------U------------
	!Convective terms    
      FUX=((U(I,J,K)+U(I+1,J,K))**2-(U(I-1,J,K)+U(I,J,K))**2)*0.25*RDX
      FUY=((V(I,J,K)+V(I+1,J,K))*(U(I,J,K)+U(I,J+1,K))-(V(I,J-1,K)+V(I+1,J-1,K))*(U(I,J-1,K)+U(I,J,K)))*0.25*RDY
      FUZ=((W(I,J,K)+W(I+1,J,K))*(U(I,J,K)+U(I,J,K+1))-(W(I,J,K-1)+W(I+1,J,K-1))*(U(I,J,K-1)+U(I,J,K)))*0.25*RDZ
	!Viscous terms
      DIFFN=U(I,J+1,K)-U(I,J,K)
      DIFFS=U(I,J,K)-U(I,J-1,K)
      DIFFE=U(I+1,J,K)-U(I,J,K)
      DIFFW=U(I,J,K)-U(I-1,J,K)
      DIFFF=U(I,J,K+1)-U(I,J,K)
      DIFFB=U(I,J,K)-U(I,J,K-1)
	! hardcode s-l BC
      IF ((FNP(I,J+1,K).EQ.0).OR.(FNP(I+1,J+1,K).EQ.0)) THEN !North border
	DIFFN=-2*U(I,J,K)
      ENDIF
      IF ((FNP(I,J-1,K).EQ.0).OR.(FNP(I+1,J-1,K).EQ.0)) THEN !South border
	DIFFS=2*U(I,J,K)
      ENDIF
      IF ((FNP(I,J,K+1).EQ.0).OR.(FNP(I+1,J,K+1).EQ.0)) THEN !Front border
	DIFFF=-2*U(I,J,K)
      ENDIF
      IF ((FNP(I,J,K-1).EQ.0).OR.(FNP(I+1,J,K-1).EQ.0)) THEN !Back border
	DIFFB=2*U(I,J,K)
      ENDIF
      VISU=(DIFFE-DIFFW)*XFAC + (DIFFN-DIFFS)*YFAC + (DIFFF-DIFFB)*ZFAC

	! Pressure forces
      DPX=P(I+1,J,K)-P(I,J,K)


    ! -------------V------------
	!Convective terms 
      FVX=((U(I,J,K)+U(I,J+1,K))*(V(I,J,K)+V(I+1,J,K))-(U(I-1,J,K)+U(I-1,J+1,K))*(V(I-1,J,K)+V(I,J,K)))*0.25*RDX
      FVY=((V(I,J,K)+V(I,J+1,K))**2-(V(I,J-1,K)+V(I,J,K))**2)*0.25*RDY 
      FVZ=((W(I,J,K)+W(I,J+1,K))*(V(I,J,K)+V(I,J,K+1))-(W(I,J,K-1)+W(I,J+1,K-1))*(V(I,J,K-1)+V(I,J,K)))*0.25*RDZ

	!Viscous terms 
      DIFFN=V(I,J+1,K)-V(I,J,K)
      DIFFS=V(I,J,K)-V(I,J-1,K)
      DIFFE=V(I+1,J,K)-V(I,J,K)
      DIFFW=V(I,J,K)-V(I-1,J,K)
      DIFFF=V(I,J,K+1)-V(I,J,K)
      DIFFB=V(I,J,K)-V(I,J,K-1)
	! hardcode s-l BC
      IF ((FNP(I+1,J,K).EQ.0).OR.(FNP(I+1,J+1,K).EQ.0)) THEN !East border
	DIFFE=-2*V(I,J,K)
      ENDIF
      IF ((FNP(I-1,J,K).EQ.0).OR.(FNP(I-1,J+1,K).EQ.0)) THEN !West border
	DIFFW=2*V(I,J,K)
      ENDIF

      IF ((FNP(I,J,K+1).EQ.0).OR.(FNP(I,J+1,K+1).EQ.0)) THEN !Front border
	DIFFF=-2*V(I,J,K)
      ENDIF
      IF ((FNP(I,J,K-1).EQ.0).OR.(FNP(I,J+1,K-1).EQ.0)) THEN !Back border
	DIFFB=2*V(I,J,K)
      ENDIF
      VISV=(DIFFE-DIFFW)*XFAC + (DIFFN-DIFFS)*YFAC + (DIFFF-DIFFB)*ZFAC

	! Pressure forces
      DPY=P(I,J+1,K)-P(I,J,K)



    ! -------------W------------
	!Convective terms  
      FWX=((U(I,J,K)+U(I,J,K+1))*(W(I,J,K)+W(I+1,J,K))-(U(I-1,J,K)+U(I-1,J,K+1))*(W(I-1,J,K)+W(I,J,K)))*0.25*RDX
      FWY=((V(I,J,K)+V(I,J,K+1))*(W(I,J,K)+W(I,J+1,K))-(V(I,J-1,K)+V(I,J-1,K+1))*(W(I,J-1,K)+W(I,J,K)))*0.25*RDY
      FWZ=((W(I,J,K)+W(I,J,K+1))**2-(W(I,J,K-1)+W(I,J,K))**2)*0.25*RDZ

	!Viscous terms 
      DIFFN=W(I,J+1,K)-W(I,J,K)
      DIFFS=W(I,J,K)-W(I,J-1,K)
      DIFFE=W(I+1,J,K)-W(I,J,K)
      DIFFW=W(I,J,K)-W(I-1,J,K)
      DIFFF=W(I,J,K+1)-W(I,J,K)
      DIFFB=W(I,J,K)-W(I,J,K-1)
	! hardcode s-l BC
      IF ((FNP(I+1,J,K).EQ.0).OR.(FNP(I+1,J,K+1).EQ.0)) THEN !East border
	DIFFE=-2*W(I,J,K)
      ENDIF
      IF ((FNP(I-1,J,K).EQ.0).OR.(FNP(I-1,J,K+1).EQ.0)) THEN !West border
	DIFFW=2*W(I,J,K)
      ENDIF
      IF ((FNP(I,J+1,K).EQ.0).OR.(FNP(I,J+1,K+1).EQ.0)) THEN !North border
	DIFFN=-2*W(I,J,K)
      ENDIF
      IF ((FNP(I,J-1,K).EQ.0).OR.(FNP(I,J-1,K+1).EQ.0)) THEN !South border
	DIFFS=2*W(I,J,K)
      ENDIF
      VISW=(DIFFE-DIFFW)*XFAC + (DIFFN-DIFFS)*YFAC + (DIFFF-DIFFB)*ZFAC

	! Pressure forces
      DPZ=P(I,J,K+1)-P(I,J,K)

	! Bouynacy forces
      FB=0.5*RA*PR*(TEMP(I,J,K)+TEMP(I,J+1,K)) ! y-dir

  




    !-------- Premliminary Velocities-------------
      IF ((FNP(I,J,K).EQ.1).OR.(FNP(I+1,J,K).EQ.1)) THEN
        DELTAU(I,J,K) = DELT*(-DPX*RDX - FUX - FUY - FUZ + PR*VISU)
      ELSE
        DELTAU(I,J,K)=0.0
        U(I,J,K)=0.0
      ENDIF

      IF ((FNP(I,J,K).EQ.1).OR.(FNP(I,J+1,K).EQ.1)) THEN
        DELTAV(I,J,K)=DELT*(-DPY*RDY - FVX - FVY - FVZ + PR*VISV + FB)
      ELSE
        DELTAV(I,J,K)=0.0
        V(I,J,K)=0.0
      ENDIF

      IF ((FNP(I,J,K).EQ.1).OR.(FNP(I,J,K+1).EQ.1)) THEN
        DELTAW(I,J,K)=DELT*(-DPZ*RDZ - FWX - FWY - FWZ + PR*VISW)
      ELSE
        DELTAW(I,J,K)=0.0
        W(I,J,K)=0.0
      ENDIF

    ENDDO
  ENDDO   
ENDDO

U(:,:,:) = U(:,:,:) + DELTAU(:,:,:)
V(:,:,:) = V(:,:,:) + DELTAV(:,:,:)
W(:,:,:) = W(:,:,:) + DELTAW(:,:,:)

   
RETURN
END

!-----------------------------------------------------------------

SUBROUTINE GEN_T_MAT 
!     purpose:
!     To construct the sparse coefficient matrix A_T for solving temperature
USE FIELD
IMPLICIT NONE
INTEGER :: N
DOUBLE PRECISION :: APO,AP,AN,AS,AE,AW,AF,AB


      
N = 0

DO I=2,IM1
  DO J=2,JM1
    DO K=2,KM1

      APO = 1.0/DELT
      AE = 0.5*U(I,J,K)/DELX - KON(I,J,K)/DELX**2 - (KON(I+1,J,K)-KON(I-1,J,K))/(4*DELX**2)
      AW = -0.5*U(I-1,J,K)/DELX - KON(I,J,K)/DELX**2 + (KON(I+1,J,K)-KON(I-1,J,K))/(4*DELX**2)
      AN = 0.5*V(I,J,K)/DELY - KON(I,J,K)/DELY**2 - (KON(I,J+1,K)-KON(I,J-1,K))/(4*DELY**2)
      AS = -0.5*V(I,J-1,K)/DELY - KON(I,J,K)/DELY**2 + (KON(I,J+1,K)-KON(I,J-1,K))/(4*DELY**2)
      AF = 0.5*W(I,J,K)/DELZ - KON(I,J,K)/DELZ**2 - (KON(I,J,K+1)-KON(I,J,K-1))/(4*DELZ**2)
      AB = -0.5*W(I,J,K-1)/DELZ - KON(I,J,K)/DELZ**2 + (KON(I,J,K+1)-KON(I,J,K-1))/(4*DELZ**2) 
      AP = 2*KON(I,J,K)*(1.0/DELX**2 + 1.0/DELY**2 + 1.0/DELZ**2)

      ! center node 
      N = N + 1
      TAVAL(N) = APO + AE + AW + AN + AS + AF + AB + 2*AP
      TI(N) = NNT(I,J,K)
      TJ(N) = NNT(I,J,K)

      !At Boundary walls, grad(T)=0
      IF (J.EQ.(JM1)) THEN !North node
        TAVAL(N) = TAVAL(N) + AN
      ENDIF
      IF (J.EQ.2) THEN !South node
        TAVAL(N) = TAVAL(N) + AS
      ENDIF
      IF (K.EQ.(KM1)) THEN !Front node
        TAVAL(N) = TAVAL(N) + AF
      ENDIF
      IF (K.EQ.2) THEN !Back node
        TAVAL(N) = TAVAL(N) + AB
      ENDIF

      !At Boundary walls, T=Known
      IF (I.EQ.(IM1)) THEN !East node
        TAVAL(N) = TAVAL(N) - AE
      ENDIF
      IF (I.EQ.2) THEN !West node
        TAVAL(N) = TAVAL(N) - AW
      ENDIF



      ! Neighbour nodes in domain
      IF(I.LT.IM1) THEN !East Node
        N = N + 1
        TAVAL(N) = AE
        Ti(N) = NNT(I,J,K)
        Tj(N) = NNT(I+1,J,K)
      ENDIF

      IF(I.GT.2) THEN !West Node
        N = N + 1
        TAVAL(N) = AW
        Ti(N) = NNT(I,J,K)
        Tj(N) = NNT(I-1,J,K)
      ENDIF

      IF(J.LT.JM1) THEN !North Node
        N = N + 1
        TAVAL(N) = AN
        Ti(N) = NNT(I,J,K)
        Tj(N) = NNT(I,J+1,K)
      ENDIF

      IF(J.GT.2) THEN !South Node
        N = N + 1
        TAVAL(N) = AS
        Ti(N) = NNT(I,J,K)
        Tj(N) = NNT(I,J-1,K)
      ENDIF

      IF(K.LT.KM1) THEN !Front Node
        N = N + 1
        TAVAL(N) = AF
        Ti(N) = NNT(I,J,K)
        Tj(N) = NNT(I,J,K+1)
      ENDIF

      IF(K.GT.2) THEN !Back Node
        N = N + 1
        TAVAL(N) = AB
        Ti(N) = NNT(I,J,K)
        Tj(N) = NNT(I,J,K-1)
      ENDIF      

      TEMP_VEC(NNT(I,J,K))=TEMP(I,J,K)

    ENDDO
  ENDDO    
ENDDO

RETURN
END

!------------------------------------------------------------
SUBROUTINE GEN_TB_VEC
!     purpose:
!     To construct b in Ax=b
USE FIELD
IMPLICIT NONE

DOUBLE PRECISION :: APO,AE,AW


TB(:)=0.0
APO = 1.0/DELT


DO I=2,IM1
  DO J=2,JM1
    DO K=2,KM1
      TB(NNT(I,J,K))= APO*TEMP_P(I,J,K)

      !At Boundary walls, T=known
      IF (I.EQ.IM1) THEN !East node
        AE = 0.5*U(I,J,K)/DELX - KON(I,J,K)/DELX**2 - (KON(I+1,J,K)-KON(I-1,J,K))/(4*DELX**2)
        TB(NNT(I,J,K)) = TB(NNT(I,J,K)) - 2*AE*TC

      ENDIF
      IF (I.EQ.2) THEN !West node
        AW = -0.5*U(I-1,J,K)/DELX - KON(I,J,K)/DELX**2 + (KON(I+1,J,K)-KON(I-1,J,K))/(4*DELX**2)
        TB(NNT(I,J,K)) = TB(NNT(I,J,K)) - 2*AW*TH
      ENDIF

    ENDDO
  ENDDO
ENDDO


RETURN
END

!------------------------------------------------------------
SUBROUTINE CORR_FS
!     purpose:
!     Correct solid fraction
USE FIELD
USE SOLVEFIELD
IMPLICIT NONE

DOUBLE PRECISION :: VAL1, VAL2, F_INV_PURE


! initial guesses for Xi
DO I=2,IM1
  DO J=2,JM1
    DO K=2,KM1
      XI(I,J,K) = FS(I,J,K) - FS_P(I,J,K) - STE*(F_INV_PURE(FS(I,J,K),TM) - TEMP(I,J,K))
      XI_N(I,J,K) = FS_N(I,J,K) - FS_P(I,J,K) - STE*(F_INV_PURE(FS_N(I,J,K), TM) - TEMP(I,J,K))
    ENDDO
  ENDDO
ENDDO


! Secant method
DO ITER_TT=1,ITMAX_TT
  ! reset residual
  RES_TT=0.0

  DO I=2,IM1
    DO J=2,JM1
      DO K=2,KM1
        ! Hold values for next iteration
        VAL1=FS(I,J,K)
        VAL2=XI(I,J,K)
      
        ! next value
        FS(I,J,K) = FS(I,J,K) - XI(I,J,K)*(FS(I,J,K)-FS_N(I,J,K))/(XI(I,J,K)-XI_N(I,J,K))
        XI(I,J,K) = FS(I,J,K) - FS_P(I,J,K) - STE*(F_INV_PURE(FS(I,J,K),TM) - TEMP(I,J,K))

        ! Replace old values
        FS_N(I,J,K) = VAL1
        XI_N(I,J,K) = VAL2

        ! residual
        RES_TT = RES_TT + ABS(XI(I,J,K)-XI_N(I,J,K))
      ENDDO
    ENDDO
  ENDDO

  ! Convergence test
  IF (RES_TT.LT.EPSI_TT) THEN
    EXIT
  ENDIF

ENDDO


! limit Solid Fraction
DO I=2,IM1
  DO J=2,JM1
    DO K=2,KM1
      FS(I,J,K) = MIN( 1.0 , MAX( 0.0 , FS(I,J,K) ) )
    ENDDO
  ENDDO
ENDDO



RETURN
END


!------------------------------------------------------------
SUBROUTINE TITER
!     Purpose: Calculate temperature field
!     at current time level
USE FIELD
USE SOLVEFIELD
IMPLICIT NONE


!Store previous time step temperature field
TEMP_P=TEMP
FS_P=FS

! second starting guess of secant method
FS_N=FS+0.1


!Generate system matrix
CALL GEN_T_MAT
  
!generate b-Vector
CALL GEN_TB_VEC

!Calculate tentative Temperature field
CALL CGSOLVER(TAVAL,TI,TJ,TB,TEMP_VEC,UCOUNT_T,ENTRIESA_T,ITER_T,RES_T,EPSI_T,ITMAX_T)

!Update tentative temperature field
DO I=2,IM1
  DO J=2,JM1
    DO K=2,KM1
      TEMP(I,J,K) = TEMP_VEC(NNT(I,J,K))
    ENDDO
  ENDDO
ENDDO

!Solid fraction Correction
CALL CORR_FS

!Correct Temperature field
DO I=2,IM1
  DO J=2,JM1
    DO K=2,KM1
      TEMP(I,J,K) = TEMP(I,J,K) + (FS(I,J,K) - FS_P(I,J,K))/STE
    ENDDO
  ENDDO
ENDDO

! Impose BC
CALL BCTEMP
  
RETURN
END

!------------------------------------------------------------
SUBROUTINE GEN_A_MAT(UNKNOWNS) 
!     purpose:
!     To construct the sparse coefficient matrix A
USE FIELD
IMPLICIT NONE

INTEGER :: UNKNOWNS,N
DOUBLE PRECISION :: AP,AN,AS,AE,AW,AF,AB


! Find entries in A so that A can be allocated
N = 0
  
DO I=2,IM1
  DO J=2,JM1
    DO K=2,KM1

      IF (FNP(I,J,K).EQ.1) THEN       	!current cell is a fluid cell
       N = N + 1       			! Add Center node

        IF(FNP(I+1,J,K).EQ.1) THEN	!fluid cell to the east
	  N = N + 1			! Add East node
    	ENDIF 
        IF(FNP(I-1,J,K).EQ.1) THEN	!fluid cell to the west
	  N = N + 1			! Add West node
    	ENDIF
        IF(FNP(I,J+1,K).EQ.1) THEN	!fluid cell to the north
	  N = N + 1			! Add North node
    	ENDIF
        IF(FNP(I,J-1,K).EQ.1) THEN	!fluid cell to the south
	  N = N + 1	  		! Add South node
	ENDIF
        IF(FNP(I,J,K+1).EQ.1) THEN	!fluid cell to the front
	  N = N + 1			! Add Front node
    	ENDIF
        IF(FNP(I,J,K-1).EQ.1) THEN	!fluid cell to the back
	  N = N + 1			! Add Back node
    	ENDIF 

      ENDIF
    ENDDO
  ENDDO
ENDDO


ENTRIESA = N
ALLOCATE(AVAL(ENTRIESA),AI(ENTRIESA),AJ(ENTRIESA))
ALLOCATE(B(UNKNOWNS),DELP(UNKNOWNS))
DELP(:) = 0.0
  
N = 0

AE = 1/(DELX**2)
AW = 1/(DELX**2)
AN = 1/(DELY**2)
AS = 1/(DELY**2)
AF = 1/(DELZ**2)
AB = 1/(DELZ**2)
AP = -(AE+AW+AN+AS+AF+AB)

DO I=2,IM1
  DO J=2,JM1
    DO K=2,KM1
      !Current node is a fluid node
      IF (FNP(I,J,K).EQ.1) THEN
        !center node 
        N = N + 1
        AVAL(N) = AP
        AI(N) = NN(I,J,K)
        AJ(N) = NN(I,J,K)

        !At Boundary wall/s-l interphase, grad(P)=0
        IF (FNP(I+1,J,K).EQ.0) THEN !East node
          Aval(N) = Aval(N) + AE
        ENDIF
        IF (FNP(I-1,J,K).EQ.0) THEN !West node
          Aval(N) = Aval(N) + AW
        ENDIF
        IF (FNP(I,J+1,K).EQ.0) THEN !North node
          Aval(N) = Aval(N) + AN
        ENDIF
        IF (FNP(I,J-1,K).EQ.0) THEN !South node
          Aval(N) = Aval(N) + AS
        ENDIF
        IF (FNP(I,J,K+1).EQ.0) THEN !Front node
          Aval(N) = Aval(N) + AF
        ENDIF
        IF (FNP(I,J,K-1).EQ.0) THEN !Back node
          Aval(N) = Aval(N) + AB
        ENDIF

        ! Neighbour nodes are fluid nodes
        IF(FNP(I+1,J,K).EQ.1) THEN !East Node
          N = N + 1
          Aval(N) = AE
          Ai(N) = NN(I,J,K)
          Aj(N) = NN(I+1,J,K)
        ENDIF

        IF(FNP(I-1,J,K).EQ.1) THEN !West Node
          N = N + 1
          Aval(N) = AW
          Ai(N) = NN(I,J,K)
          Aj(N) = NN(I-1,J,K)
        ENDIF

        IF(FNP(I,J+1,K).EQ.1) THEN !North Node
          N = N + 1
          Aval(N) = AN
          Ai(N) = NN(I,J,K)
          Aj(N) = NN(I,J+1,K)  
        ENDIF

        IF(FNP(I,J-1,K).EQ.1) THEN !South Node
          N = N + 1
          Aval(N) = AS
          Ai(N) = NN(I,J,K)
          Aj(N) = NN(I,J-1,K)
        ENDIF

        IF(FNP(I,J,K+1).EQ.1) THEN !Front Node
          N = N + 1
          Aval(N) = AF
          Ai(N) = NN(I,J,K)
          Aj(N) = NN(I,J,K+1)  
        ENDIF

        IF(FNP(I,J,K-1).EQ.1) THEN !Back Node
          N = N + 1
          Aval(N) = AB
          Ai(N) = NN(I,J,K)
          Aj(N) = NN(I,J,K-1)
        ENDIF


      ENDIF
    ENDDO
  ENDDO
ENDDO    

PREFNODE = 1  ! Setting the node at which the pressure will always be zero

  DO N=1,ENTRIESA		! Setting reference pressure to zero
    IF((Ai(N).EQ.PREFNODE).AND.(Aj(N).NE.PREFNODE)) THEN
      Aval(N) = 0.0
    ENDIF
    IF((Ai(N).EQ.PREFNODE).AND.(Aj(N).EQ.PREFNODE)) THEN
      Aval(N) = -1.0*10**8
    ENDIF
  ENDDO     

RETURN
END

!--------------------------------------------------------------------------

SUBROUTINE GEN_B_VEC
!     purpose:
!     To construct b in Ax=b
USE FIELD
IMPLICIT NONE

DOUBLE PRECISION :: DIV
INTEGER :: N

N=0

DO I=2,IM1
  DO J=2,JM1
    DO K=2,KM1

      ! Current node is a fluid node
      IF (FNP(I,J,K).EQ.1) THEN
        N=N+1
        DIV = RDX*(U(I,J,K)-U(I-1,J,K)) + RDY*(V(I,J,K)-V(I,J-1,K)) + RDZ*(W(I,J,K)-W(I,J,K-1))
        B(NN(I,J,K))=DIV/(DELT)
      ENDIF

    ENDDO
  ENDDO
ENDDO


B(PREFNODE) = 0.0	! Assigning reference pressure 0 to pressure reference cell

RETURN
END

!------------------------------------------------------------
SUBROUTINE PITER
!     Purpose:To solve the Poissons equation of pressure
!     and adjust velocity and pressure fields so that 
!     continuity is maintained for every time step.
USE FIELD
USE SOLVEFIELD
IMPLICIT NONE


!Calculate tentative velocity field
CALL VEL

! impose BC
CALL BCVEL


! main Momentum loop (not running if there are no fluid nodes)
IF (UCOUNT.GT.0) THEN

  !Generate A matrix in Ax=b
  CALL GEN_A_MAT(UCOUNT)	

  !Generate b vector in Ax=b
  CALL GEN_B_VEC
		
  !Solve system
  CALL CGSOLVER(AVAL,AI,AJ,B,DELP,UCOUNT,ENTRIESA,ITER_P,RES_P,EPSI_P,ITMAX_P) 


  DO I=2,IM1-1	! Updating U-components, if fluid node
    DO J=2,JM1
      DO K=2,KM1
        IF((FNP(I+1,J,K).EQ.1).AND.(FNP(I,J,K).EQ.1)) THEN
          U(I,J,K)=U(I,J,K) + DELT*(DELP(NN(I,J,K)) - DELP(NN(I+1,J,K)))/DELX
        ENDIF
      ENDDO
    ENDDO
  ENDDO

  DO I=2,IM1		! Updating V-components, if fluid node
    DO J=2,JM1-1
      DO K=2,KM1
        IF((FNP(I,J+1,K).EQ.1).AND.(FNP(I,J,K).EQ.1)) THEN
          V(I,J,K)=V(I,J,K) + DELT*(DELP(NN(I,J,K)) - DELP(NN(I,J+1,K)) )/DELY
        ENDIF
      ENDDO
    ENDDO
  ENDDO   

  DO I=2,IM1		! Updating W-components, if fluid node
    DO J=2,JM1
      DO K=2,KM1-1
        IF((FNP(I,J,K+1).EQ.1).AND.(FNP(I,J,K).EQ.1)) THEN
          W(I,J,K)=W(I,J,K) + DELT*(DELP(NN(I,J,K)) - DELP(NN(I,J,K+1)) )/DELZ
        ENDIF
      ENDDO
    ENDDO
  ENDDO   


  DO I=2,IM1		! Updating pressure field, if fluid node
    DO J=2,JM1
      DO K=2,KM1
        IF (FNP(I,J,K).EQ.1) THEN
          P(I,J,K) = P(I,J,K) + DELP(NN(I,J,K))
        ENDIF
      ENDDO
    ENDDO
  ENDDO   
  
  ! impose BC
  CALL BCVEL

  ! Deallocate (size is changing)
  DEALLOCATE(AVAL,AI,AJ)
  DEALLOCATE(B,DELP)

ENDIF
RETURN
END 

!------------------------------------------------------------------------------------


! ========= CONJUGATE GRADIENT SOLVER =========


SUBROUTINE CGSOLVER(Aval,Ai,Aj,b,x,unknowns,entriesA, iter, res,epsi, itmax)	! Solves the system Ax=b with the conjugate gradients method

  IMPLICIT NONE
  INTEGER :: unknowns, entriesA,itmax,iter
  DOUBLE PRECISION, DIMENSION (entriesA) :: Aval
  INTEGER, DIMENSION (entriesA) :: Ai, Aj
  DOUBLE PRECISION, DIMENSION (unknowns) :: tempvec, x, r, d, b
                               
  DOUBLE PRECISION :: alpha, beta, tempscal1, tempscal2,epsi,res
  


  iter=0

  x(:) = 0.0
  r(:) = b(:)
  d(:) = r(:) 
  res = sum(abs(r))
  IF(res.LT.epsi) THEN		! Convergence test
     RETURN
  ENDIF


  DO iter=1,itmax
    CALL VECVEC(r,r,tempscal1,unknowns)			! (r'*r)
    CALL MATVEC(Aval,Ai,Aj,d,tempvec,unknowns,entriesA)	! (A*d)
    CALL VECVEC(d,tempvec,tempscal2,unknowns)		! (d'*A*d)
    alpha = tempscal1/tempscal2				! alpha = (r'*r)/(d'*A*d)
    x(:) = x(:) + alpha*d(:)				! x = x + alpha*d
    tempvec(:) = r(:) - alpha*tempvec(:)   		! r_new = r - alpha*A*d
    CALL VECVEC(tempvec,tempvec,tempscal2,unknowns)	! (r_new'*r_new)
    beta = tempscal2/tempscal1				! beta = (r_new'*r_new)/(r'*r)
    r(:) = tempvec(:)					! r = r_new
    d(:) = r(:) + beta*d(:)  				! d = r + beta*d

    res = sum(abs(r))
    IF(res.LT.epsi) THEN		! Convergence test
      EXIT
    ENDIF

  ENDDO
  

RETURN
END


!------------------------------------------------------------------------------------



SUBROUTINE MATVEC(Aval,Ai,Aj,u,x,m,n)	! Matrix-vector multiplication, x=Au  -- A: matrix, u: vector, x: return vector, m: length of u, n: entries in A
  IMPLICIT NONE
  INTEGER :: i, m, n
  DOUBLE PRECISION, DIMENSION(n) :: Aval 
  INTEGER, DIMENSION(n) :: Ai, Aj 
  DOUBLE PRECISION, DIMENSION(m) :: u,x
  
  x(:) = 0
  DO i=1,n
    x(Ai(i)) = x(Ai(i)) + Aval(i)*u(Aj(i))
  ENDDO
  
RETURN
END

SUBROUTINE MATTVEC(Aval,Ai,Aj,u,x,m,n)	! Transposed matrix-vector multiplication, x=Au  -- A: matrix, u: vector, x: return vector, m: length of u, n: entries in A
  IMPLICIT NONE
  INTEGER :: i, m, n
  DOUBLE PRECISION, DIMENSION(n) :: Aval 
  INTEGER, DIMENSION(n) :: Ai, Aj 
  DOUBLE PRECISION, DIMENSION(m) :: u,x
  
  x(:) = 0
  DO i=1,n
    x(Aj(i)) = x(Aj(i)) + Aval(i)*u(Ai(i))
  ENDDO
  
RETURN
END


SUBROUTINE VECVEC(u,v,x,m)	! Vector-vector multiplication, x=uv -- u: vector1, v: vector2, x: return scalar, m: size of u and v
  IMPLICIT NONE
  INTEGER :: i, m
  DOUBLE PRECISION, DIMENSION(m) :: u, v
  DOUBLE PRECISION :: x
  
  x = 0.0
  DO i=1,m
    x = x + u(i)*v(i)
  ENDDO
  
RETURN
END




!------------------------------------------------------------

SUBROUTINE WRITE_COMP_DAT
!     purpose:
!     To create a dat file for use in tecplot
!     Will be used to create an animation
USE FIELD
IMPLICIT NONE

! Write to INTERPHASE file
! EXPERIMENT "Ben-David et.al. (2011)"



WRITE(INTERP,*)'ZONE ','T="E, t=120"',' J=',15,', DATAPACKING=POINT'
WRITE(INTERP,*)0.139742319,' ',0.000000000
WRITE(INTERP,*)0.141228940,' ',0.071165644
WRITE(INTERP,*)0.141228940,' ',0.142331288
WRITE(INTERP,*)0.141228940,' ',0.213496933
WRITE(INTERP,*)0.141228940,' ',0.285889571
WRITE(INTERP,*)0.141228940,' ',0.355828221
WRITE(INTERP,*)0.141228940,' ',0.428220859
WRITE(INTERP,*)0.142715560,' ',0.500613497
WRITE(INTERP,*)0.145688801,' ',0.570552147
WRITE(INTERP,*)0.147175421,' ',0.642944785
WRITE(INTERP,*)0.150148662,' ',0.714110429
WRITE(INTERP,*)0.153121903,' ',0.785276074
WRITE(INTERP,*)0.153121903,' ',0.856441718
WRITE(INTERP,*)0.153121903,' ',0.927607362
WRITE(INTERP,*)0.153121903,' ',1.000000000
WRITE(INTERP,*)
WRITE(INTERP,*)'ZONE ','T="E, t=300"',' J=',15,', DATAPACKING=POINT'
WRITE(INTERP,*)0.266105055,' ',0.000000000
WRITE(INTERP,*)0.266105055,' ',0.071165644
WRITE(INTERP,*)0.266105055,' ',0.142331288
WRITE(INTERP,*)0.266105055,' ',0.213496933
WRITE(INTERP,*)0.269078295,' ',0.285889571
WRITE(INTERP,*)0.269078295,' ',0.355828221
WRITE(INTERP,*)0.272051536,' ',0.428220859
WRITE(INTERP,*)0.276511397,' ',0.500613497
WRITE(INTERP,*)0.282457879,' ',0.570552147
WRITE(INTERP,*)0.291377602,' ',0.642944785
WRITE(INTERP,*)0.301783944,' ',0.714110429
WRITE(INTERP,*)0.315163528,' ',0.785276074
WRITE(INTERP,*)0.327056492,' ',0.856441718
WRITE(INTERP,*)0.334489594,' ',0.927607362
WRITE(INTERP,*)0.334489594,' ',1.000000000
WRITE(INTERP,*)
WRITE(INTERP,*)'ZONE ','T="E, t=600"',' J=',15,', DATAPACKING=POINT'
WRITE(INTERP,*)0.325569871,' ',	0.000000000
WRITE(INTERP,*)0.325569871,' ',0.071165644
WRITE(INTERP,*)0.327056492,' ',0.142331288
WRITE(INTERP,*)0.328543112,' ',0.213496933
WRITE(INTERP,*)0.337462834,' ',0.285889571
WRITE(INTERP,*)0.344895937,' ',0.355828221
WRITE(INTERP,*)0.355302279,' ',0.428220859
WRITE(INTERP,*)0.382061447,' ',0.500613497
WRITE(INTERP,*)0.422200198,' ',0.570552147
WRITE(INTERP,*)0.477205154,' ',0.642944785
WRITE(INTERP,*)0.530723489,' ',0.714110429
WRITE(INTERP,*)0.566402379,' ',0.785276074
WRITE(INTERP,*)0.581268583,' ',0.856441718
WRITE(INTERP,*)0.585728444,' ',0.927607362
WRITE(INTERP,*)0.585728444,' ',1.000000000
WRITE(INTERP,*)
WRITE(INTERP,*)'ZONE ','T="E, t=1000"',' J=',15,', DATAPACKING=POINT'
WRITE(INTERP,*)0.393954410,' ',0.000000000
WRITE(INTERP,*)0.395441031,' ',0.071165644
WRITE(INTERP,*)0.401387512,' ',0.142331288
WRITE(INTERP,*)0.411793855,' ',0.213496933
WRITE(INTERP,*)0.443012884,' ',0.285889571
WRITE(INTERP,*)0.520317146,' ',0.355828221
WRITE(INTERP,*)0.611000991,' ',0.428220859
WRITE(INTERP,*)0.686818632,' ',0.500613497
WRITE(INTERP,*)0.741823588,' ',0.570552147
WRITE(INTERP,*)0.784935580,' ',0.642944785
WRITE(INTERP,*)0.814667988,' ',0.714110429
WRITE(INTERP,*)0.833994054,' ',0.785276074
WRITE(INTERP,*)0.841427156,' ',0.856441718
WRITE(INTERP,*)0.842913776,' ',0.927607362
WRITE(INTERP,*)0.842913776,' ',1.000000000
WRITE(INTERP,*)


RETURN
END 


